/*
VierGewinnt -- Learning Java
Copyright (C) <2006>  Robin and Detlef Wolf
This library is free software -- GNU LGPL license applies.
see http://www.gnu.org/licenses/lgpl.txt for details
*/


public class VierGewinntField {
  // holds the playing field
  // this is the only piece of code that knows how to make a move and check whether it is winning;
  // outside read access to columnNum, rowNum and field is fine;
  // read access to field from the outside is granted because of performance reasons.

  public static final int columnNum = 7 ;
  public static final int rowNum = 5 ;
  private static final int winLength = 4 ;

  public int field[][] = new int[rowNum][columnNum] ; // 0=empty, 1=Player1, 2=Player2
         // row 0 is bottom row
         // row rowNum-1 is top row

  // put() and unPut() could be made considerably faster if there were int columnRow[0..columnNum-1]
  // in which the row index of the bottommost free field is kept;
  // any volunteer?

  public VierGewinntField() {
    initialize() ; 
  }

  public void initialize() {
    for (int row=0; row<rowNum; ++row)
      for (int col=0; col<columnNum; ++col) 
        field[row][col] = 0 ;
  }

  public boolean put(int column, int who) {
    // column -- [0..columnNum-1]
    // who    -- 1=Player1, 2=Player2
    // return true if move possible, else false
    for (int row=0; row<rowNum; ++row)
      if (field[row][column] == 0) {
        field[row][column] = who ;
        return true ;
      }
    return false ;
  }

  public boolean putPossible(int column) {
    // column -- spalte [0..columnNum-1]
    // return true if move possible, else false
    return field[rowNum-1][column] == 0 ;
  }

  public boolean isTie() {  // true if no more move possible
    for (int col=0; col<columnNum; ++col) 
      if (putPossible(col))
        return false ;
    return true ;
  }

  public void unPut(int column) {
    // inverse of put()
    // column --  [0..columnNum-1]
    for (int row=rowNum-1; row>=0; --row)
      if (field[row][column] != 0) {
        field[row][column] = 0 ;
        break ;
      }
  }

  // direction vectors row/column
  private static final int dr[] = {1, 1, 0,  1} ;
  private static final int dc[] = {0, 1, 1, -1} ;
  
  public boolean isWinningPos(int row, int col, int who) {
    // return true if filling (row,col) by 'who' would be a winning move
    // else false
    int r ;    // current row
    int c ;    // current column
    int der ;  // delta row
    int dec ;  // delta column
    for (int dir=0; dir<dr.length; ++dir) {
      int cnt = 1 ;
      der = dr[dir] ;
      dec = dc[dir] ;
      r = row ;
      c = col ;
      for (int len=1; len<winLength; ++len) {
        r += der ;
        c += dec ;
        if (r < 0 || r>=rowNum ||
            c < 0 || c>=columnNum ||
            field[r][c] != who) 
          break ;
        ++cnt ;
      }
      if (cnt >= winLength) // if we are lucky the new move was at the end of chain
        return true ;
      // otherwise add opposite direction
      der = -der ;
      dec = -dec ;
      r = row ;
      c = col ;
      for (int len=1; len<winLength; ++len) {
        r += der ;
        c += dec ;
        if (r < 0 || r>=rowNum ||
            c < 0 || c>=columnNum ||
            field[r][c] != who) 
          break ;
        ++cnt ;
      }
      if (cnt >= winLength) 
        return true ;
    }
    return false ;
  }


  public static final int NOT_POSSIBLE = 0 ;
  public static final int OK_NO_WIN = 1 ;
  public static final int OK_IS_WIN = 2 ;

  public int putCheckWin(int column, int who) {
    // try to put stone of player 'who' into 'column' and check whether this was a winning move
    // column -- spalte [0..columnNum-1]
    // who    -- wer? 1=Player1, 2=Player2
    // return 0 = put not possible, 1=put done, no win, 2=put done and win 
    if (putPossible(column))  // little performance optimization if column is full
      for (int row=0; row<rowNum; ++row)
        if (field[row][column] == 0) {
          field[row][column] = who ;
          return isWinningPos(row, column, who) ? OK_IS_WIN : OK_NO_WIN ; 
        }
    return NOT_POSSIBLE ;
  }

  public int probeCheckWin(int column, int who) {
    // same as putCheckWin() except that it doesn't put
    if (putPossible(column))  // little performance optimization if column is full
      for (int row=0; row<rowNum; ++row)
        if (field[row][column] == 0) 
          return isWinningPos(row, column, who) ? OK_IS_WIN : OK_NO_WIN ; 
    return NOT_POSSIBLE ;
  }


}

